"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const _ = require("lodash");
let AMQP = require('amqplib/callback_api');
/**
 * The AMPQ feature uplifts local chassis events to the global event bus.
 *
 * It bind the AMQP service to this.amqp upon successful connection.
 *
 * @type {{usecase: [string], connect: module.exports.connect, publish_message: module.exports.publish_message, fn: module.exports.fn}}
 */
class amqp {
    constructor() {
        this.name = "amqp";
        this.title = "Transmit local events to AMQP topics";
        this.amqp = {};
        /**
         * AMQP
         * ----
         * Connect to the AMQP message broker
         *
         * @param context
         * @param options
         * @param callback
         */
        this.connect = function (context, options, callback) {
            let user_pass = options.username ? options.username + ":" + options.password + "@" : "";
            let url = options.protocol + "://" + user_pass + "" + options.host + (options.port ? ":" + options.port : "");
            context.log({ code: "api:amqp:connecting", message: "connecting to AMQP", url: url });
            // if we're already connected - assume we're done
            if (this.amqp) {
                context.warn({ code: "api:amqp:connected", message: "connecting to AMQP", url: url });
                callback(false, { code: "connected", url: url });
                return;
            }
            // login to the message broker
            AMQP.connect(url, function (err, conn) {
                if (err) {
                    this.amqp = false;
                    err = { code: "api:amqp:connect:failed", error: err, url: url };
                    context.error(err);
                    callback(err, false);
                    return;
                }
                this.amqp = { connection: conn };
                callback(false, { code: "connected", url: url });
            });
        };
        /**
         * Send a message to a queue
         *
         * @param context
         * @param options
         * @param remote_queue
         * @param message
         * @param callback
         */
        this.publish_message = function (context, options, remote_queue, message, callback) {
            assert(options, "missing options");
            assert(this.amqp && this.amqp.connection, "not connected");
            if (!this.amqp) {
                callback({ code: "api:amqp:disconnected", remote_queue: remote_queue, error: "misconfigured" });
                return;
            }
            this.amqp.connection.createChannel(function (err, ch) {
                if (err) {
                    err = { code: "api:amqp:message:failed", error: err, remote: remote_queue };
                    context.error(err);
                    callback(err, false);
                    return;
                }
                ch.assertQueue(remote_queue, { durable: false });
                ch.sendToQueue(remote_queue, new Buffer(JSON.stringify(message)), callback);
                let log_msg = { code: "api:amqp:message:send", remote: remote_queue, size: message.length };
                context.log(log_msg);
            });
        };
    }
    /**
     * bridge local NodeJS events and the global AMQP message bus
     *
     * @param context
     * @param options
     * @returns {exports}
     */
    install(context, options) {
        assert(context, "missing context");
        assert(context.api, "missing context.api");
        assert(context.config, "missing context.config");
        let config = context.config;
        // default AMQP configuration
        let exchange = options.exchange || options.queue || config.name;
        let AMQP_options = _.extend({ protocol: "AMQP", host: "localhost", port: 5672, vhost: "/", username: "", password: "", queue: exchange }, config.amqp, options);
        options = _.extend({ on: { "api:started": true } }, options);
        let self = this;
        // initial connect
        this.connect(context, AMQP_options, function (err, msg) {
            // handle errors
            if (err) {
                context.log({ code: "api:ampqp:error:connect", message: "connection failed", error: err });
                return;
            }
            context.log({ code: "api:ampqp:connected", message: "connected", error: err });
            // uplift/republish local NodeJS events as global MQ events
            _.each(options.on, function (remote_queue, local_event) {
                // if event queue is "true" then use the default queue
                if (remote_queue === true)
                    remote_queue = exchange;
                if (remote_queue) {
                    context.log({ code: "api:ampqp:listen", message: "listen", source: local_event, target: remote_queue });
                    // relay the local Node event to AMQP pub/sub queue
                    context.bus.on(local_event, function (data) {
                        let payload = { event: local_event, timestamp: Date.now(), source: local_event, target: remote_queue, data: null };
                        // log it - WITHOUT the data
                        let local_log = _.extend({}, payload, { code: "api:ampqp:republish", message: "republish" });
                        // unpack nested .data
                        if (data) {
                            payload.data = data.data || data;
                        }
                        self.publish_message(context, options, remote_queue, payload, null);
                        context.log(local_log);
                    });
                }
            });
            // clean-up connection before exiting
            process.once('SIGINT', function () {
                this.amqp.connection.close.bind(this.amqp.connection);
                context.log({ code: "api:ampqp:disconnected", message: "connected", error: err, msg_url: msg.url });
            });
        });
    }
}
exports.default = amqp;
//# sourceMappingURL=amqp.js.map