"use strict";
/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2019 Troven Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const _ = require("lodash");
// import apidocs from "./../middleware/apidocs";
// import {IChassisPlugin, IChassisContext} from "../interfaces";
// const path = require("path");
const request = require("request");
const exphbs = require('express-handlebars');
const Handlebars = require("handlebars");
const templateSettings = {
    interpolate: /\{\{(.+?)\}\}/g
};
/**
 * View CMS Page
 * -------------
 * A simple "micro-CMS" that renders a Handlebars page based on static meta-data and optional dynamic API call.
 *
 * The same JWT used for this UI layer will be forwarded to the API endpoint (unless reuseJWT = true).
 *
 * Bindings can be injected from config or options -
 *      static:
 *      brand:
 *          title: Site / App name
 *          description: Generic site description
 *      apis:
 *          my_app:
 *              url: request-url
 *              headers: request-headers
 *              etc
 *      feature:
 *          a: feature toggle A
 *          b: toggle B
 *
 * @type {{name: string, title: string, description: string, _render: module.exports._render, fn: module.exports.fn}}
 */
class views {
    constructor() {
        this.name = "views";
        this.title = "simple server-side Handlebars rendering";
        this._helpers = function () {
            Handlebars.registerHelper('json', function (data) {
                return JSON.stringify(data);
            });
        };
        // render a dynamic page
        this._render = function (context, options) {
            let config = context.config;
            let views_apis = _.extend({}, config.apis, options.apis);
            this._helpers();
            return function (req, res) {
                let page_properties = {
                    static: _.extend({}, config.static, options.static),
                    brand: _.extend({}, config.brand, options.brand),
                    i18n: _.extend({}, config.i18n, options.i18n),
                    apis: _.extend({}, config.apis, options.apis)
                };
                //            console.log("PAGE: %o", page_properties.openapi);
                // determine the layout and page paths
                let layout = req.params.layout || "";
                let page = req.params[0] || req.path || "index";
                let page_path = (layout ? layout + "/" : "") + page;
                let path_parts = page_path.split("/");
                // setup page-level variables
                let page_data = _.extend({ page: { title: page } }, page_properties);
                let page_context = _.extend({ data: page_data, layout: layout || "default", errors: [] }, page_data);
                // lookup the backend API from the page path, page or layout
                let api = null, template_path = null;
                for (let i = path_parts.length; i > 0; i--) {
                    let _path = path_parts.slice(0, i).join("/");
                    if (!api) {
                        template_path = _path;
                        api = views_apis[_path];
                    }
                }
                // if we found a matching API ...
                if (api && api.url) {
                    // transform API (template) into concrete URI
                    let _api_path_context = _.extend({}, path_parts);
                    let resolve_api_path = _.template(api.url, templateSettings)(_api_path_context);
                    // console.log("render url: %s ", new_api);
                    // retrieve API response and inject into the page context
                    let field_name = api.field || "views";
                    page_context.data[field_name] = {};
                    let _headers = {};
                    // inject Bearer Token
                    if (options.reuseJWT !== false) {
                        _headers.authorization = req.header("authorization");
                    }
                    // call backend API
                    request({ uri: resolve_api_path, headers: _headers }, function (err, httpResponse, body) {
                        assert(err && httpResponse, "invalid views response");
                        // if an parsing error occurs, add error to page_context - but continue rendering
                        try {
                            page_context[field_name] = JSON.parse(body);
                        }
                        catch (e) {
                            console.error("remote detail: %s --> %j", e, body);
                            page_context.errors.push({ message: e });
                        }
                        context.log({ code: "views:api", message: "render API", render: template_path, page: page, url: resolve_api_path });
                        // render our page
                        res.render(template_path, page_context);
                    });
                }
                else {
                    // display a simple dynamic page
                    context.log({ code: "views:page", message: "render page", render: page_path, page: page });
                    res.render(page_path, page_context);
                }
            };
        };
    }
    install(context, _options) {
        let options = _options;
        // let config = context.config;
        assert(options.mountPath, "missing options.mountPath");
        // template rendering options
        let render_options = _.extend({
            defaultLayout: 'default',
            layoutsDir: "views/layouts",
            partialsDir: "views/partials",
            extname: ".hbs"
        }, options.render);
        // create the Handlebars renderer
        let hbs = exphbs.create(render_options);
        assert(hbs.engine, "Invalid handlebars");
        // attach render to engine
        context.api.engine('.hbs', hbs.engine);
        context.api.set('view engine', '.hbs');
        // site-wide properties
        // set up the routes
        let base_path = options.mountPath || "";
        context.api.get(base_path + "/:layout/*.html", this._render(context, options));
        context.api.get(base_path + "/*.html", this._render(context, options));
        // redirect home
        context.api.get(base_path + "/", function (req, res) {
            assert(req && req, "invalid middleware");
            res.redirect(base_path + "/index.html");
        });
        context.log({ code: "views:published", message: "views published", path: base_path });
    }
}
exports.default = views;
//# sourceMappingURL=views.js.map