

import * as _ from "lodash";
import {IChassisContext, IChassisPlugin, IChassisMiddleware, IOperation, APIs } from "api-service-core";
import assert = require("assert");
import passport = require("passport");
import { Strategy } from "passport-keycloak-oauth2-oidc";
import cookie_parser = require('cookie-parser');
import * as express_session from 'express-session';
import * as jwt from "jsonwebtoken";

/**
 * plugin to handle JSON body
 *
 * @param context
 * @param options
 * @returns {exports}
 */
export default class KeycloakPlugin implements IChassisPlugin {

    name = "keycloak";
    title = "Inject passport plugins before  request is processed";

    kc_options: any = {};

    constructor() {
    }

    install(context: IChassisContext, _options: any) {

        assert(passport, "missing passport");
        // Register the strategy with passport

        this.kc_options = _.defaults({}, context.config.keycloak, _options, {
            authServerURL: process.env.KEYCLOAK_SERVER,
            realm: process.env.KEYCLOAK_REALM,
            publicClient: false,
            sslRequired: 'external',
            clientID: process.env.KEYCLOAK_CLIENT_ID,
            clientSecret: process.env.KEYCLOAK_CLIENT_SECRET,
            strategy: "keycloak"
        });

        let strategy = this.kc_options.strategy;
        assert(strategy, "missing keycloak strategy");
        this.kc_options.callback = this.kc_options.callback || "/callback/"+strategy;
        this.kc_options.logout = this.kc_options.logout || "/logout";

        this.kc_options.callbackURL = this.kc_options.callbackURL || this.kc_options.callback;
        this.kc_options.tokenURL = this.kc_options.tokenURL || this.kc_options.authServerURL+"/realms/"+this.kc_options.realm+"/protocol/openid-connect/auth"
        this.kc_options.userInfoURL = this.kc_options.userInfoURL || this.kc_options.authServerURL+"/realms/"+this.kc_options.realm+"/protocol/openid-connect/auth"
        this.kc_options.logoutURL = this.kc_options.logoutURL || this.kc_options.authServerURL+"/realms/"+this.kc_options.realm+"/protocol/openid-connect/logout"

        context.api.use(cookie_parser());

        let session_options = _.extend({
            secret: strategy+"_secret_",
            resave: true,
            saveUninitialized: true
        }, context.config.session, _options);

        context.api.use( express_session(session_options) );
        context.api.use(passport.initialize());
        context.api.use(passport.session());

        passport.serializeUser(function(user, done) {
            console.log("serializeUser: %j", user);
            done(null, user);
          });
          
        passport.deserializeUser(function(user, done) {
            console.log("deserializeUser: %j", user);
            done(null, user);
        });

        passport.use( strategy, new Strategy(this.kc_options, 
            (accessToken, _refreshToken, profile, done) => {
                context.log( { code: "plugin:keycloak:profile", profile: profile});
                let decoded = jwt.decode(accessToken);
                _.extend(profile, decoded);
                delete profile._json;
                // profile.accessToken = accessToken;
                // profile.refreshToken = refreshToken;
                done(null, profile);
            })
        );

        context.middleware.set( new KeycloakMiddleware(this));

        // setup authentication callback endpoint

        let callback_options = _.extend({
            failureFlash : false,
            successRedirect: "/",
            failureRedirect: this.kc_options.logoutURL,
        }, _options);

        context.api.get( this.kc_options.callback, (req,res,next) => {
            context.log({ code: "plugin:keycloak:callback" });
            passport.authenticate(strategy, callback_options)(req,res,next);
        } );

        context.api.get( this.kc_options.logout, (_req,_res) => {
            context.log({ code: "plugin:keycloak:logout" });
            APIs.get(context, this.kc_options.logoutURL).then( (res) =>{
                console.log("LOGGED OUT: %j", res.statusCode);
                _res.redirect("/");
            }).catch( (err) => {
                _res.status(500);
                _res.send({ code: "logout:failed ", error: err});
            })
        });

        context.log( { code: "plugin:keycloak:installed", options: this.kc_options});
    }
}

class KeycloakMiddleware implements IChassisMiddleware {
    name = "keycloak";

    constructor(protected plugin: KeycloakPlugin) {
    }

    fn(_operation: IOperation, _options: any): Function {

        let context = _operation.context
        let options = _.extend({}, _options);
        let strategy = this.plugin.kc_options.strategy;

        return ( (req, res, next) => {
            if (!req.isAuthenticated() || !req.isAuthenticated) {  
                if (req.session) {  
                    req.session.redirectUrl = req.headers.referer || req.originalUrl || req.url;  
                }  
                context.log( { code: "plugin:keycloak:not-authenticated", path: req.path, session: req.session, strategy: strategy, redirectUrl: req.session.redirectUrl });

            } else {
                context.log( { code: "plugin:keycloak:authenticated", path: req.path, strategy: strategy });
            }
            return passport.authenticate(strategy, options)(req,res,next);
        });
    }
}
