

import * as _ from "lodash";
import {IChassisContext, IChassisPlugin, IChassisMiddleware, IOperation, APIs, Vars, Utils, JWT } from "api-service-core";
import axios from 'axios'
import * as qs from 'qs'
import * as jwt from "jsonwebtoken";

const GRANT_UMA_TICKET = 'urn:ietf:params:oauth:grant-type:uma-ticket'

/**
 * plugin to handle JSON body
 *
 * @param context
 * @param options
 * @returns {exports}
 */
export default class UMA implements IChassisPlugin {

    name = "uma";
    title = "Check UMA rules before request is processed";

    uma_options: any = {};
    uma_discovery: any = {};
    context: IChassisContext;
    uma_pat: any = { expires_in: 60 };
    DEBUG = true
    OPTIONS_WHITELIST = [ "clientID", "realm", "authServerURL"]

    constructor() {
    }

    install(context: IChassisContext, _options: any) {
        this.context = context;
        // Register the strategy with passport

        this.uma_options = _.defaults({}, context.config.keycloak, _options, {
            authServerURL: process.env.KEYCLOAK_SERVER,
            realm: process.env.KEYCLOAK_REALM,
            clientID: process.env.KEYCLOAK_CLIENT_ID,
            clientSecret: process.env.KEYCLOAK_CLIENT_SECRET
        });

        this.uma_options.realmURL = this.uma_options.authServerURL+"/realms/"+this.uma_options.realm
        this.uma_options.discoveryURL = this.uma_options.realmURL+"/.well-known/uma2-configuration"

        // retrieve discovery document from Keycloak
        this.discover_endpoints(context)

        context.middleware.set( new UMAMiddleware(this));
        context.log( { code: "plugin:uma:installed", options: this.uma_options});
    }

    discover_endpoints(context: IChassisContext) {
        APIs.request(context,this.uma_options.discoveryURL, {}).then( (disco) => {
            this.uma_discovery = disco;
            context.log( { code: "plugin:uma:discovered", discovery: disco });

            this.retrieve_PAT(this.context);
    
        }).catch( err => {throw new Error(err) })


    }

    retrieve_PAT(context: IChassisContext) {
        let url = this.uma_discovery.token_endpoint;

        return this.post_form(url, {}, { grant_type: 'client_credentials', client_id: this.uma_options.clientID , client_secret: this.uma_options.clientSecret })
        .then( (token=> {
            this.uma_pat = token
            this.DEBUG && context.log( { code: "plugin:uma:token", message: "expires_in "+(token?token.expires_in+" seconds":false) });
            setTimeout( () => this.retrieve_PAT(this.context), 900 * token.expires_in );
        })).catch( err=> {
            context.error( { code: "plugin:uma:token:error", message: err });
            setTimeout( () => this.retrieve_PAT(this.context), 10000 );
        })


    }

    post_form( url: string, headers: any, fields: any ): Promise<any> {
        let context = this.context;
        headers['Content-Type'] = 'application/x-www-form-urlencoded'
        let form = qs.stringify(fields)

        // console.log("form: %j -> %j", fields, form.toString());
        return new Promise(( resolve, reject ) => {
            // call backend API
            this.DEBUG && context.log({ code: "uma:request:post", url: url, headers: headers, data: form });

            axios({ url: url, method: 'post', headers: headers, data: form }).then( res => {
                if (res.status>=400) return reject("failed "+res.status)
                resolve(res.data)
            }).catch( err => {
                this.DEBUG && console.log("form-error: %j", err); 
                reject(err.message) 
            } )
       });
    }

    retrieveEntitlements(user_token: string, resource: string) {
        let headers = { 'Authorization': 'Bearer '+user_token };
        // this.uma_options.entitlementsURL
        let endpoint = this.uma_options.entitlementsURL+this.uma_options.clientID;
        this.context.log( { code: "plugin:uma:entitlements", resource: resource, endpoint: endpoint });
        return this.post_form(endpoint, headers, {})
    }

    retrievePermissions(user_token: string, resource: string, scope: string, _params?: any): Promise<any> {
        let headers = { 'Authorization': 'Bearer '+user_token };
        let permission = resource+"#"+scope;
        let params = _.extend({ grant_type: GRANT_UMA_TICKET, audience: this.uma_options.clientID, permission: permission }, _params)

        this.context.log( { code: "plugin:uma:permissions", resource: resource, scope: scope, jwt: user_token?true:false, token_endpoint: this.uma_discovery.token_endpoint });
        return this.post_form(this.uma_discovery.token_endpoint, headers, params)
    }

    getRSID(uri: string): Promise<any> {
        let headers = { 'Authorization': 'Bearer '+this.uma_pat.access_token}
        let rego_url = this.uma_discovery.resource_registration_endpoint;
        return axios( { url: rego_url, method: 'get', headers: headers, params: { uri: uri } })
    }

    /**
     * Retrieve a Permission ticket:
     * 
     *  1. lookup the RSID from the URI
     *  2. map the RSID + SCOPE into claims
     *  3. request ticket for the claims
     * 
     * @param uri 
     * @param scope 
     */
    retrievePermissionTicket(uri: string, scope: string): Promise<any> {
        if (!this.uma_pat.access_token) return Promise.reject("missing-pat");

        let headers = { 'Authorization': 'Bearer '+this.uma_pat.access_token}
        let rego_url = this.uma_discovery.resource_registration_endpoint;
        let perm_url = this.uma_discovery.permission_endpoint; // "http://localhost:9000" // 
        this.context.log( { code: "plugin:uma:query_resources", headers: headers, url: uri });
        
        return new Promise( (resolve, reject) => {

            // 1. lookup resource set (RSID) for resource URI
            axios( { url: rego_url, method: 'get', headers: headers, params: { uri: uri } }).then( rs_body => {

                // 2. map claims
                let claims = _.map(rs_body.data, (rs => {
                    return { resource_id: rs, resource_scopes: [scope] }
                }))

                this.DEBUG && console.log("rs.claims: %j --> %j", perm_url, claims)

                // 3. retreive permission ticket for RSID & scope
                axios( { url: perm_url, method: 'post', headers: headers, data: claims }).then( r => resolve(r.data) ).catch( err => { 
                    this.DEBUG && console.log("rs.claims.failed: %j", err.message)
                    reject(err.message)
                 } )
            }).catch( err => { 
                this.DEBUG && console.log("rs.map.failed: %s -> %j", err.response.status, err.response.data)
                reject(err.message)
             } )
        })
    }

    /**
     * Send a formatted `WWW-Authenticate` response 
     * @param res 
     * @param ticket 
     */
    sendPermissionTicket(res: any, ticket: string) {
        res.set('WWW-Authenticate', "UMA realm=\""+this.uma_options.realm+"\", as_uri=\""+this.uma_options.realmURL+"\", ticket=\""+ticket+"\"")
        res.status(401)
        res.send()
    }

    /**
     * 
     * Retrieve a user's RPT
     * 
     * @param user_jwt 
     * @param ticket 
     */
    retrieveRPT(user_jwt: string, ticket:string ): Promise<any> {
        let url = this.uma_discovery.token_endpoint
        let headers = JWT.getHeaders(user_jwt)
        let fields = { grant_type: GRANT_UMA_TICKET, ticket: ticket }
        this.DEBUG && console.log("retrieveRPT.post: %j -> %j", headers, fields)
        return new Promise( (resolve, reject) => {
            this.post_form( url, headers, fields ).then( r => resolve(r) ).catch( err => { 
                this.DEBUG && console.log("retrieveRPT.failed: %j", err.message)
                reject(err.message)
            })
        })
    }

    /**
     * retrieve RPT
     * 
     * 1. retrievePermissionTicket
     * 2. retrieveRPT
     * 
     * @param user_token 
     * @param resource 
     * @param scope 
     */
    retrievePermissionRPT(user_token: string, resource: string, scope: string) {
        return new Promise( (resolve, reject) => {
            this.retrievePermissionTicket( resource, scope ).then( data=> {
                this.DEBUG && console.log("permissions.ok: %j --> %j", resource, jwt.decode(data.ticket))
    
                this.retrieveRPT(user_token, data.ticket).then( resolve ).catch( reject )
            }).catch(reject)
        });
    }

            // this.plugin.retrievePermissionTicket( resource, options.scope ).then( data=> {
            //     this.DEBUG && console.log("permissions.ok: %j --> %j", resource, jwt.decode(data.ticket))

            //     this.plugin.retrieveRPT(user_token, data.ticket).then( (rpt: any) => {
            //         this.DEBUG && console.log("rpt.ok: %j", rpt)
            //         next();
            //     }).catch( (err)=> {
            //         this.DEBUG && console.error("rpt.failed: %j", err.toString())
            //     })

            //     // if (!user_token) {
            //     //     this.plugin.sendPermissionTicket(res, data.ticket)
            //     // } else {
            //     //     next();
            //     // }
            // }).catch(err => {
            //     this.DEBUG && console.error("permission.failed: %j", err.toString())
            //     Utils.sendError(context, res, {statusCode: 403, code: "api:oauth:forbidden", message: "access-denied"});
            // })
}

class UMAMiddleware implements IChassisMiddleware {
    name = "uma";

    constructor(protected plugin: UMA) {
    }

    fn(_operation: IOperation, _options: any): Function {

        let context = _operation.context

        return ( (req, res, next) => {

            let user_token = JWT.getBearerToken(context,req);
            // if (!JWT.getValidJWT(context,req,res)) return;

            // get resource context
            let req_ctx = APIs.getOperationContext(_operation, req, { plugin: _.pick(this.plugin.uma_options,this.plugin.OPTIONS_WHITELIST) } );

            let resource = req.path
            let options = Vars.$$( _.extend({}, _options), req_ctx );
            let scope = options.scope || _operation.operationId

            context.log( { code: "api:uma:request", resource: resource, scope: scope });

            this.plugin.retrievePermissionRPT(user_token, resource, scope).then( (rpt) => {
                this.plugin.DEBUG && console.error("retrievePermissionRPT.ok: %j", rpt)
                next()
            }).catch(err =>{
                this.plugin.DEBUG && console.error("retrievePermissionRPT.failed: %s", err)
                Utils.sendError(context, res, {statusCode: 403, code: "api:oauth:forbidden", message: "access-denied"});
            })


        });
    }
}
