

import * as _ from "lodash";
import * as Prometheus from "prom-client";
import { IOperation, IChassisContext, IChassisPlugin, IChassisMetrics, IChassisMiddleware, Operation } from "api-service-core";
import { Request, Response } from "express";

/**
 * plugin to handle prometheus metrics. 
 * The plugin installs "metrics" middleware that can be attached to a endpoint to return collected statistics.
 *
 * @param context
 * @param options
 * @returns {exports}
 */
export default class prometheus implements IChassisPlugin {

    name = "prometheus";
    title = "support for Prometheus metrics";

    install(context: IChassisContext, options: any) {

        false && options;

        const collectDefaultMetrics = Prometheus.collectDefaultMetrics;

        // Collect default metrics every 5 seconds
        collectDefaultMetrics({ timeout: 5000 });

        // Inject Prometheus instrumentation as IChassisMetrics
        let metrics: IChassisMetrics = new PrometheusMetrics(context, options);
        context.metrics = metrics;

        let labels = _.extend({}, options.labels, {
            serviceId: context.pkg.name,
            serviceVersion: context.pkg.version,
            serviceUUID: context.uuid
        });
        Prometheus.register.setDefaultLabels( labels );

        metrics.metric("chassis_get_all_metrics", "Counter");
        context.middleware.set(new PrometheusMetricsMiddleware());
        
        // count everything that is audited
        context.bus.on("api:audit", (_audit)=> {
            if (!_audit.code) return;

            let labels = { 
                serviceId: context.uuid, 
                serviceName: context.config.name, 
                serviceVersion: context.pkg.version
            };
            metrics.metric(_audit.code, "Counter", labels);
            metrics.count(_audit.code, 1, labels);
        });

    }
}

class PrometheusMetricsMiddleware implements IChassisMiddleware {
    name:string = "api.metrics";

    fn(operation: IOperation, options: any): Function {
        false && options;
        false && operation;

        return (req: Request, res: Response) => {
            false && req;
            res.status(200);

            operation.context.metrics.count("chassis_get_all_metrics", 1);
            res.send( Prometheus.register.metrics() );
        };
    }
}

class PrometheusMetrics implements IChassisMetrics {
    metrics: any = {};
    labels: any = {};

    constructor(protected context: IChassisContext, protected options: any) {
        false && context;
        false && options;
        this.labels = options.labels || {};
    }

    simple_key(key: string): string {
        return key.replace(/[\W_]+/g,"_");
    }

    _metric(metric_key): any {
        return this.metrics[ this.simple_key(metric_key) ]
    }

    _labels(oper?: any) {
        let labels = _.extend({}, {
            serviceId: this.context.pkg.name,
            serviceVersion: this.context.pkg.version,
            serviceUUID: this.context.uuid
        }, this.options.labels);
        if (oper && oper.operationId) {
            labels.operationId = oper.operationId;
            labels.resource = oper.resource || "unknown";
            labels.actionId = oper.actionId || "unknown";
        }
        return labels;
    }

    metric ( metric_key: string, type?: string, labels?: any): any {
        let key = this.simple_key(metric_key);
        if (this.metrics[key]) return this.metrics[key];

        type = type || "Counter";
        labels = this._labels(labels);
        let labelNames = Array.isArray(labels)?labels:_.keys(labels);
        let options = { name: key, help: type+" for "+metric_key , labelNames: labelNames };
        let fn = new Prometheus[type](options);
        this.metrics[key] = fn;
        return fn;
    }

    count ( metric_key: string, amount?: number, oper?: Operation): void {
        let labels = this._labels(oper);
        this.metric(metric_key, "Counter", labels);
        this._metric(metric_key).inc(labels, amount);
    }

    value ( metric_key: string, value: number, oper?: Operation): void {
        let labels = this._labels(oper);
        this._metric(metric_key).set(labels, value);
    }

    history (metric_key: string, value: number, oper?: Operation): void {
        let labels = this._labels(oper);
        this._metric(metric_key).observe(labels, value);
    }

}