/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */


import * as _ from "lodash";

import {IChassisContext, IChassisPlugin, IChassisMiddleware, IOperation, APIs, APIRequests, Vars }  from "api-service-core";

const request = require("request");
const exphbs  = require('express-handlebars');
const Handlebars = require("handlebars");

/**
 * View CMS Page
 * -------------
 * A simple "micro-CMS" that renders a Handlebars page based on static meta-data and optional dynamic API call.
 *
 * The same JWT used for this UI layer will be forwarded to the API endpoint (unless reuseJWT = true).
 *
 * Bindings can be injected from config or options -
 *      static:
 *      brand:
 *          title: Site / App name
 *          description: Generic site description
 *      apis:
 *          my_app:
 *              url: request-url
 *              headers: request-headers
 *              etc
 *      feature:
 *          a: feature toggle A
 *          b: toggle B
 *
 * @type {{name: string, title: string, description: string, _render: module.exports._render, fn: module.exports.fn}}
 */
export default class render implements IChassisPlugin {

    name = "render";
    title = "Convenient server-side rendering";

    render_options: any;
    hbs: any;

    install(context: IChassisContext, _options: any) {

        let options = _options as any;

        // template rendering options

        this.render_options = _.extend({
            defaultLayout: 'default',
            layoutsDir: "views/layouts",
            partialsDir: "views/partials",
            extname: ".hbs"
        }, options );

        // create the Handlebars renderer

        this.hbs = exphbs.create(this.render_options);

        // attach render to engine

        context.api.engine(".hbs", this.hbs.engine);
        context.api.set('view engine', '.hbs');

        // site-wide properties

        Handlebars.registerHelper('json', function(data: any) {
            return JSON.stringify(data);
        });

        Handlebars.registerHelper('html', function( text: string) {
            let replaced = text?text.replace(new RegExp('\r?\n','g'), '<br/>'):"";
            return replaced;
        });

        context.middleware.set( new RenderView(context, _options) );
    }
}

// interface IViewRequest {

// }

class RenderView implements IChassisMiddleware {
    name = "api.render";
    context: IChassisContext;
    options: any;

    constructor(context: IChassisContext, options: any) {
        this.context = context;
        this.options = options;
    }

    /**
     * Collect a set of API responses as promise
     * 
     * @param apis 
     * @param options 
     * @param token 
     */

    _requests(apis: any, options: any, token: string): Promise<any> {
        let _requests = [];

        // iterate over set of APIs
        for (let api_name in apis) {
            let api = apis[api_name];

            let _headers: any = {};
            if (token) {
                // inject Bearer Token
                _headers.authorization = token;
            }
            let api_url = Vars.$(api.url, options);

            let _request = this.backend_request( api_name, api_url, _headers);
            _requests.push( _request );
        }

        return Promise.all(_requests);
    }

    backend_request( api_name: string, url: string, _headers: any ): Promise<any> {
        let context = this.context;

        return new Promise(( resolve, reject ) => {
            // call backend API
            context.log({ code: "plugin:render:api", api: api_name, url: url, headers: _headers });

            request({ uri: url, headers: _headers }, function (err: any, httpResponse: any, body: any) {
                if (err || !httpResponse) {
                    context.error({ code: "plugin:render:api:failed", api: api_name, url: url, httpResponse: httpResponse?true:false, error: err?err.toString():"no error" });
                    reject(err);
                } else {
                    try {
                        let result = {};
                        result[api_name] = JSON.parse(body);
                        resolve( result );
                    } catch (e) {
                        reject(e);
                    }
                }
            });
        });
    }

    fn(operation: IOperation, _options: any): Function {

        let options = _.extend({}, this.options, _options);
        let context = operation.context;
        let feature_apis: APIRequests = _.extend({}, this.options.apis, options.apis );
        // let is_static = (_.isEmpty(feature_apis));

        // console.log("OPTIONS: %j", _options);

        let page_defaults = {
            brand: _.extend({}, this.options.brand, options.brand),
            i18n: _.extend({}, this.options.i18n, options.i18n),
            pkg: { name: context.pkg.name, version: context.pkg.version }
        };

        context.log({ code: "plugin:render", resource: operation.resource,reuseJWT: options.reuseJWT, page: page_defaults, feature_apis: _.keys(feature_apis), pkg: page_defaults.pkg });

        // renderer middleware
        return (req, res) => {

            let jwt_token = options.reuseJWT?req.header("authorization"):"";

            let req_params = _.extend({}, {}, options.params, req.params );

            // parameters for .hbs render
            let page_context = _.extend({}, options.render, { params: req_params, query: req.query, url: req.url } , page_defaults);

            // prefer 'page' from request, then options, otherwise 'index'

            let page = req.params.page || req.params[0] || options.page || "index";

            let ix = page.lastIndexOf(".");
            if (ix>0) {
                page = page.substring(0,ix);
            }
            page = page + ".hbs"

            page_context.params.page = page;
            let page_path = (page_context.params.path?page_context.params.path+"/":"")+page;
            page_context.layout = options.layout || page_context.params.layout || "default";

            // console.log("RENDER: (%s) %s -> %j -> %j", page_context.layout, page_path, req.params, page_context);

            // parameters for API requests
            let api_params = _.extend( {}, options.params, req_params, page_context );

            // get the API responses ...

            context.log({ code: "plugin:render:apis", apis: _.keys(feature_apis), params: api_params });

            APIs.requests(this.context, feature_apis, api_params, jwt_token).then(results=> {
                // merge API responses
                page_context = _.extend(page_context, results);

                // render page response
                try {
                    res.render(page_path, page_context);
                    context.log({ code: "plugin:render:page:active", page: page_path, layout: page_context.layout, context: _.keys(page_context) });
                } catch(e) {
                    res.status(404);
                    res.send({ code: "plugin:render:page:active", error: e.message });
                }

            }).catch(err => {
                // render error response
                let _error = { code: "plugin:render:page:active:failed", page: page_path, layout: page_context.layout, message: err.message, error: err.toString() };
                context.error(_error);
                res.status(500);
                res.send(_error);
            });

        }
    }
}
