/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */

// import * as assert from "assert";
import * as _ from "lodash";
import {IChassisContext, IChassisPlugin} from "api-service-core";
import axios from 'axios'

/**
 * Web Hooks
 * ---------
 * Feature connects selected local events to web hooks that notify a set of HTTP listeners.
 *
 * When the event is triggered, each listener receives an HTTP POST with a payload like:
 *  {
 *      code:  "some:operation",
 *      params: query params
 *      payload: req.body
 *  }
 *
 * @type {{name: string, title: string, _notify_listeners: module.exports._notify_listeners, fn: module.exports.fn}}
 */
export default class webhooks implements IChassisPlugin {
    name = "webhooks";
    title = "Trigger requests from local events";

    LOW_SUCCESS_CODE: 200;
    HIGH_SUCCESS_CODE: 299;

    _notify_listeners = function(context, flow, event_id, hook_id, event, options) {

        // resolve flow
        flow = _.extend({ id: hook_id, title: hook_id, listeners:[] },flow);
        let default_cmd = _.extend({ method: "post", headers: {}, strictSSL: true }, flow.defaults, options.request );

        context.log( { code: "api:webhooks:trigger", message: "webhook triggered", hook: hook_id, event: event_id } );

        // notify each listener in turn ..

        _.each(flow.listeners, function(listener) {

            // build the HTTP request command
            let cmd = _.extend({}, default_cmd, listener);
            cmd.json = { payload: event, now: Date.now(), hook: hook_id, trigger: event_id };

            context.log( { code: "api:webhook", message: "call webhook", hook: hook_id, event: event_id, url: cmd.url, method: cmd.method } );

            // attempt to notify listener ...
            axios(cmd).then( res => { //, body
                if (res.status >= this.LOW_SUCCESS_CODE && res.status <= this.HIGH_SUCCESS_CODE) {
                    // request was successful
                    context.log( { code: "api:webhooks:sent", message: "webhook issued", hook: hook_id, event: event_id, url: cmd.url, method: cmd.method } );
                    context.audit({ code: "api:webhooks:sent", hook: hook_id, event: event_id, url: cmd.url, method: cmd.method });
                } else {
                    // remote endpoint returned an error
                    context.logger.error( { code: "api:webhooks:", message: "webhook rejected", status: res.status, hook: hook_id, event: event_id, url: cmd.url, method: cmd.method } );
                }
            }).catch(err => {
                context.logger.error( { code: "api:webhooks:error", message: "webhook failed", hook: hook_id, event: event_id, error: err, url: cmd.url, method: cmd.method } );
            })
        });
    };

    install(context: IChassisContext, _options: any) {

        let options = _options as any;

        // resolve hooks from file or API
        let basePath = options.mount || "/webooks";

        // build flow map
        let hooks = {};
        context.log( { code: "api:webhooks", message: "webhook registered", options: options } );

        _.each(options.hooks, function(f) { hooks[f.id] = f; });

        // listen for named events
        _.each(options.on, function(hook_id, event_id) {
            context.log( { code: "api:webhooks:register", message: "webhook registered", hook: hook_id, event: event_id } );
            // notify listeners event_id even occurs
            context.bus.on(event_id, function(event) {
                this._notify_listeners(context, hooks[hook_id], event_id, hook_id, event, options);
            });
        });

        // canary - callback endpoint
        context.api.use(basePath+"/canary/:source", function(req,res) {
            context.audit({ code: "api:webhooks:canary:"+req.params.source, params: req.params, payload: req.body });
            res.send({ now: Date.now(), params: req.params });
        });

        // canary - emit endpoint
        context.api.use(basePath+"/trigger/:event", function(req,res) {
            var emit_event = { code: "api:webhooks:trigger:"+req.params.event, params: req.params, payload: req.body };
            context.bus.emit(emit_event.code, emit_event);
            res.send(emit_event);
        });

        context.log( { code: "api:webhooks:init", message: "webhook registered", hooks: _.keys(hooks) } );
    }
};
